/* atomic-x86_64.S */

/*
 * Copyright 2007, 2008, 2009, 2010, 2011, 2012 by
 *	University of New Hampshire Interoperability Laboratory
 *	121 Technology Drive, Suite 2
 *	Durham, NH 03824-4716
 *	phone: 1-603-862-0090
 *	fax: 1-603-862-4181
 *	e-mail: info@iol.unh.edu
 * All rights reserved.

 * This file is part of the UNH EXS implementation of the Consortium
 * Specification: Extended Sockets API (ES-API) Issue 1.0, Document
 * Number C050 published by The Open Group, January 2005, and copyright
 * 2005 by The Open Group.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING.GPL in the main directory of this source tree, or the
 * BSD license, available from the file COPYING.BSD in the main directory
 * of this source tree.
 *
 * This material is based upon work supported by the National Science
 * Foundation under Grant No. OCI-1127228.
 *
 * Any opinions, findings, and conclusions or recommendations
 * expressed in this material are those of the authors and do not
 * necessarily reflect the views of the National Science Foundation.
 */

/*
 * unsigned int atomic_add_reset(unsigned int increment,
 *              unsigned int limit, volatile unsigned int *result)
 *
 * atomically updates the value in *result to
 *   *result + increment         if that is < limit
 *   0                           otherwise
 *
 * Note that the reset only applies to *result; the returned value
 * will not be reset.
 *
 * on 64-bit machines
 * %edi  unsigned int increment
 * %esi  unsigned int limit
 * %rdx  *result
 * %ebx  stores new result without reset temporarily
 * %eax  return unsigned int (new result without reset)
 *
 * free use of all registers
 */

	.text
	.align	4
	.globl	atomic_add_reset

atomic_add_reset:
	pushq	%rbx		# Push old value of rbx register (for ebx use)
	movl	(%rdx),%eax	# get old result into eax for 1st time thru loop
1:
	movl	%eax,%ecx	# get current result into ecx
	addl	%edi,%ecx	# add increment to current result in ecx
	movl	%ecx,%ebx	# return new result (without reset) in ebx
	cmpl	%ecx,%esi	# compare result + increment in ecx with limit
	cmovbel	zero(%rip),%ecx	# reset ecx to 0 if result + increment >= limit
	lock			# makes the cmpxchgl atomic
	cmpxchgl %ecx,(%rdx)	# change result to ecx if current result == eax
	jnz	1b		# loop if result changed by another task, and
				# thus eax has been reloaded with that result
	# fall thru if new result = old result + increment
	movl	%ebx,%eax	# return new result (without reset) in eax
	popq	%rbx		# Restore old value of rbx register
	ret			# return what the caller produced in eax

zero:	.long	0		# constant for reseting result

/*
 the cmpxchgl src,dest works atomically as follows:
	eax is implicitly referenced and is expected to hold old dest value
	src must be a register and should hold the new dest value
	dest can be memory or a register
 the idea is to replace old dest value with new dest value IF dest is unchanged

 when eax == dest then			# old dest value unchanged since eax
	ZF = 1; dest = src		# so replace old dest with new dest
					# old dest value remains in eax
 otherwise when eax != dest then	# old dest value changed since eax
	ZF = 0; eax = dest		# get current (changed) dest into eax

 the cmovccl src,dest works as follows:
	when the EFLAGS match the cc conditions (as in the Jcc instruction)
	then the move takes place, else it doesn't
 in the EFLAGS condition codes:
	"above" and "below" refer to the relation between unsigned integers
	"greater" and "less" refer to the relation between signed integers
 Note that the following common instructions do NOT change the EFLAGS:
	lea, mov, push, pop
 */

/*
 * unsigned int atomic_add(int increment, volatile unsigned int *result)
 *
 * atomically updates the value in *result to
 *   *result + increment
 *
 * on 64-bit machines
 * %edi  unsigned int increment
 * %rsi  *result
 * %eax  return unsigned int (new result)
 *
 * free use of all registers
 */

	.text
	.align	4
	.globl	atomic_add

atomic_add:
	movl	(%rsi),%eax	# get old result into eax for 1st time thru loop
1:
	movl	%eax,%ecx	# get current result into ecx
	addl	%edi,%ecx	# add increment to current result in ecx
	lock			# makes the cmpxchgl atomic
	cmpxchgl %ecx,(%rsi)	# change result to ecx if current result == eax
	jnz	1b		# loop if result changed by another task, and
				# thus eax has been reloaded with that result
	# fall thru if new result = old result + increment
	movl	%ecx,%eax	# return new result in eax
	ret			# return what the caller produced in eax

/*
 unsigned int	atomic_set_bits(unsigned int mask,
				volatile unsigned int *result)
 unsigned int	atomic_clear_bits(unsigned int mask,
				volatile unsigned int *result)

 on 64-bit machines
	%edi	unsigned int mask
	%rsi	*result
	%eax	return unsigned int (new result)

 free use of all registers
 */

	.text
	.align	4
	.globl	atomic_set_bits

atomic_set_bits:
	movl	(%rsi),%eax	# get old result into eax
1:
	movl	%eax,%edx	# copy current result into edx
	orl	%edi,%edx	# or mask onto current result
	lock
	cmpxchgl %edx,(%rsi)	# try to change current result
	jnz	1b		# try again if change didn't happen, and
				# eax has been reloaded with current result
	# fall thru if new result = old result | mask
	movl	%edx,%eax	# return new result in eax
	ret			# so caller knows what he produced

	.text
	.align	4
	.globl	atomic_clear_bits

atomic_clear_bits:
	notl	%edi		# ~mask into edi
	movl	(%rsi),%eax	# get old result into eax
1:
	movl	%eax,%edx	# copy current result into edx
	andl	%edi,%edx	# and ~mask onto current result
	lock
	cmpxchgl %edx,(%rsi)	# try to change current result
	jnz	1b		# try again if change didn't happen, and
				# eax has been reloaded with current result
	# fall thru if new result = old result & ~mask
	movl	%edx,%eax	# return new result in eax
	ret			# so caller knows what he produced

/*
	unsigned int atomic_swap_if_match(
				unsigned int swap_value,
				unsigned int match_value,
				volatile unsigned int *result)

	atomically updates the value in *result to
	*swap_value		if match_value == match_value

	equivalent to:
		atomically:
			unsigned int ret = *result;
			iffff (*result == match_value) {
				*result = swap_value;
			}
			return ret;

		returns value of *result prior to attempted swap
			so return value == match_value if swap was done,
						and now *result == swap_value
			   return value == *result if swap was not done,
						and now *result == old result


 on 64-bit machines
	%edi	unsigned int swap_value
	%esi	unsigned int match_value
	%rdx	*result
	%eax	return unsigned int (old result)

 free use of all registers
 */

	.text
	.align	4
	.globl	atomic_swap_if_match

atomic_swap_if_match:
	movl	%esi,%eax	# get match_value into eax
	lock			# makes the cmpxchgl atomic
	cmpxchgl %edi,(%rdx)	# change result to edi if current result == eax
/*
	always fall thru
 iffff change didn't happen because %eax was not equal to (%rdx)
	then %eax has been reloaded with old result (%rdx)
 iffff change did happen because %eax was equal to (%rdx)
	then new result = swap_value
	and neither %eax nor %edi were changed, only (%rdx) was changed
	so %eax already has old result because that equaled the match_value
 */
	ret

/*
 the cmpxchgl src,dest works atomically as follows:
	eax is implicitly referenced and is expected to hold old dest value
	src must be a register and should hold the new dest value
	dest can be memory or a register
 the idea is to replace old dest value with new dest value IF dest is unchanged

 when eax == dest then			# old dest value unchanged since eax
	ZF = 1; dest = src		# so replace old dest with new dest
					# old dest value remains in eax
 otherwise when eax != dest then	# old dest value changed since eax
	ZF = 0; eax = dest		# get current (changed) dest into eax
*/

	.text
	.align	4
	.globl	atomic_swap_if_match_ptr
atomic_swap_if_match_ptr:
	movq	%rsi,%rax	# get match_value into rax
	lock			# makes the cmpxchgq atomic
	cmpxchgq %rdi,(%rdx)	# change result to rdi if current result == rax
	ret

/* Ensure that stack is non-executable */
#if defined(__linux__) && defined(__ELF__)
.section .note.GNU-stack,"",%progbits
#endif
